/***************************************************************************
 *   Copyright (C) 2007-2013 Mike Kinghan, imk@burroingroingjoing.com      *
 *   All rights reserved.                                                  *
 *                                                                         *
 *   Contributed originally by Mike Kinghan, imk@burroingroingjoing.com    *
 *                                                                         *
 *   Redistribution and use in source and binary forms, with or without    *
 *   modification, are permitted provided that the following conditions    *
 *   are met:                                                              *
 *                                                                         *
 *   Redistributions of source code must retain the above copyright        *
 *   notice, this list of conditions and the following disclaimer.         *
 *                                                                         *
 *   Redistributions in binary form must reproduce the above copyright     *
 *   notice, this list of conditions and the following disclaimer in the   *
 *   documentation and/or other materials provided with the distribution.  *
 *                                                                         *
 *   Neither the name of Symbian Software Ltd. nor the names of its        *
 *   contributors may be used to endorse or promote products derived from  *
 *   this software without specific prior written permission.              *
 *                                                                         *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS   *
 *   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT     *
 *   LIMITED TO, THE IMPLIED WARRANTIES OF  MERCHANTABILITY AND FITNESS    *
 *   FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE        *
 *   COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,   *
 *   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,  *
 *   BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS *
 *   OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED    *
 *   AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,*
 *   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF *
 *   THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH  *
 *   DAMAGE.                                                               *
 *                                                                         *
 **************************************************************************/
#include "directive.h"
#include "expression_parser.h"
#include "diagnostic.h"
#include "lexicon.h"
#include "line_edit.h"
#include "io.h"
#include "chew.h"
#include "line_select.h"
#include "hash_include.h"
#include "if_control.h"
#include "contradiction.h"
#include "canonical_string.h"
#include <iostream>

/*!
	\file directive.cpp
	This file implements specializations of class template `directive<Type>`
*/

using namespace std;

//! \cond NO_DOXYGEN

template<> std::string const directive<HASH_IF>::_keyword_ = TOK_IF;
template<> std::string const directive<HASH_IFDEF>::_keyword_ = TOK_IFDEF;
template<> std::string const directive<HASH_IFNDEF>::_keyword_  = TOK_IFNDEF;
template<> std::string const directive<HASH_ELSE>::_keyword_  = TOK_ELSE;
template<> std::string const directive<HASH_ELIF>::_keyword_  = TOK_ELIF;
template<> std::string const directive<HASH_ENDIF>::_keyword_  = TOK_ENDIF;
template<> std::string const directive<HASH_DEFINE>::_keyword_  = TOK_DEFINE;
template<> std::string const directive<HASH_UNDEF>::_keyword_  = TOK_UNDEF;
template<> std::string const directive<HASH_INCLUDE>::_keyword_  = TOK_INCLUDE;
template<> std::string const directive<HASH_PRAGMA>::_keyword_  = TOK_PRAGMA;
template<> std::string const directive<HASH_ERROR>::_keyword_  = TOK_ERROR;

template<>
map<string,bool> directive<HASH_IF>::_directives_tab_ = map<string,bool>();
template<>
map<string,bool> directive<HASH_IFDEF>::_directives_tab_ = map<string,bool>();
template<>
map<string,bool> directive<HASH_IFNDEF>::_directives_tab_ = map<string,bool>();
template<>
map<string,bool> directive<HASH_ELSE>::_directives_tab_ = map<string,bool>();
template<>
map<string,bool> directive<HASH_ELIF>::_directives_tab_ = map<string,bool>();
template<>
map<string,bool> directive<HASH_ENDIF>::_directives_tab_ = map<string,bool>();
template<>
map<string,bool> directive<HASH_DEFINE>::_directives_tab_ = map<string,bool>();
template<>
map<string,bool> directive<HASH_UNDEF>::_directives_tab_ = map<string,bool>();
template<>
map<string,bool> directive<HASH_INCLUDE>::_directives_tab_ = map<string,bool>();
template<>
map<string,bool> directive<HASH_PRAGMA>::_directives_tab_ = map<string,bool>();
template<>
map<string,bool> directive<HASH_ERROR>::_directives_tab_ = map<string,bool>();

//! \endcond NO_DOXYGEN

#if CXX11_HAVE_UNIFORM_INITIALIZATION
vector<directive_base::evaluator> directive_base::_evaluator_tab_ = {
	&directive<HASH_UNKNOWN>::eval,
	&directive<HASH_IF>::eval,
	&directive<HASH_IFDEF>::eval,
	&directive<HASH_IFNDEF>::eval,
	&directive<HASH_ELSE>::eval,
	&directive<HASH_ELIF>::eval,
	&directive<HASH_ENDIF>::eval,
	&directive<HASH_DEFINE>::eval,
	&directive<HASH_UNDEF>::eval,
	&directive<HASH_INCLUDE>::eval,
	&directive<HASH_PRAGMA>::eval,
	&directive<HASH_ERROR>::eval
};
#else
directive_base::evaluator directive_base::_evaluator_tab_initor_[] =  {
	&directive<HASH_UNKNOWN>::eval,
	&directive<HASH_IF>::eval,
	&directive<HASH_IFDEF>::eval,
	&directive<HASH_IFNDEF>::eval,
	&directive<HASH_ELSE>::eval,
	&directive<HASH_ELIF>::eval,
	&directive<HASH_ENDIF>::eval,
	&directive<HASH_DEFINE>::eval,
	&directive<HASH_UNDEF>::eval,
	&directive<HASH_INCLUDE>::eval,
	&directive<HASH_PRAGMA>::eval,
	&directive<HASH_ERROR>::eval
};
size_t directive_base::_n_evaluators_ =
    sizeof(directive_base::_evaluator_tab_initor_)/
		sizeof(directive_base::evaluator);
vector<directive_base::evaluator>
directive_base::_evaluator_tab_(&directive_base::_evaluator_tab_initor_[0],
	&directive_base::_evaluator_tab_initor_[directive_base::_n_evaluators_]);
#endif

#if CXX11_HAVE_UNIFORM_INITIALIZATION
map<string, directive_type>
directive_base::_keyword_to_type_map_ = {
	{	string(TOK_IF),         HASH_IF },
	{	string(TOK_IFDEF),      HASH_IFDEF },
	{	string(TOK_IFNDEF),     HASH_IFNDEF	 },
	{	string(TOK_ELSE),       HASH_ELSE },
	{	string(TOK_ELIF),       HASH_ELIF },
	{	string(TOK_ENDIF),      HASH_ENDIF },
	{	string(TOK_DEFINE),     HASH_DEFINE },
	{	string(TOK_UNDEF),      HASH_UNDEF },
	{	string(TOK_INCLUDE),    HASH_INCLUDE },
	{	string(TOK_PRAGMA),     HASH_PRAGMA },
	{	string(TOK_ERROR),      HASH_ERROR }
};
#else
directive_base::keyword_map_element
directive_base::_keyword_to_type_map_initor_[] = {
	keyword_map_element(string(TOK_IF),HASH_IF),
	keyword_map_element(string(TOK_IFDEF),HASH_IFDEF),
	keyword_map_element(string(TOK_IFNDEF),HASH_IFNDEF),
	keyword_map_element(string(TOK_ELSE),HASH_ELSE),
	keyword_map_element(string(TOK_ELIF),HASH_ELIF),
	keyword_map_element(string(TOK_ENDIF),HASH_ENDIF),
	keyword_map_element(string(TOK_DEFINE),HASH_DEFINE),
	keyword_map_element(string(TOK_UNDEF),HASH_UNDEF),
	keyword_map_element(string(TOK_INCLUDE),HASH_INCLUDE),
	keyword_map_element(string(TOK_PRAGMA),HASH_PRAGMA),
	keyword_map_element(string(TOK_ERROR),HASH_ERROR)
};
size_t directive_base::_n_keywords_ =
    sizeof(directive_base::_keyword_to_type_map_initor_)/
		sizeof(directive_base::keyword_map_element);
map<string, directive_type>
directive_base::
_keyword_to_type_map_(&directive_base::_keyword_to_type_map_initor_[0],
	&directive_base::_keyword_to_type_map_initor_[directive_base::_n_keywords_]);

#endif

directive_type directive_base::get_type(std::string const & keyword)
{
#if CXX11_HAVE_AUTO_DECL
	auto where = _keyword_to_type_map_.find(keyword);
#else
	std::map<std::string, directive_type>::const_iterator where = 
				_keyword_to_type_map_.find(keyword);
#endif
	if (where == _keyword_to_type_map_.end()) {
		warning_unknown_directive() << "Unknown directive #"
		                            << keyword << emit();
		chewer_base::set_line_state(chewer_base::LS_CODE);
		return HASH_UNKNOWN;
	}
	return where->second;
}

void directive_base::report(std::string const & keyword,
                            std::string const & arg)
{
	std::cout << '#' << keyword << ' ' << arg;
	if (options::list_location()) {
		std::cout << ": " << io::in_file_name() << '(' << line_edit::line_num() << ')';
	}
	std::cout << '\n';
}

void directive_base::report(bool seen,
                            std::string const & keyword,
                            std::string const & arg)
{
	if (!seen) {
		report(keyword,arg);
	} else if (!options::list_only_once()) {
		report(keyword,arg);
	}
}

line_type
directive_base::eval_ifdef_or_ifndef(	directive_type type, char const **ptext)
{
	chewer<discardable> chew;
	char const * cp = *ptext;
	cp = chew(cp);
	symbol sym(symbol::read(cp,&cp));
	line_type retval = LT_IF;
	if (sym.configured()) {
		if (line_select::eligible(CMD_SYMBOLS)) {
			sym.report();
		}
		if (sym.defined()) {
			/* symbol is -Ded*/
			retval = type == HASH_IFDEF ?
			         LT_TRUE : LT_FALSE;
		} else {
			/* symbol is -Ued*/
			retval = type == HASH_IFNDEF ?
			         LT_TRUE : LT_FALSE;
		}
	} else {
		if (line_select::eligible(CMD_SYMBOLS)) {
			sym.record(UNCONFIGURED);
			sym.report();
		}
		if (options::implicit()) {
			retval = type == HASH_IFDEF ?
			         LT_FALSE : LT_TRUE;
		}
	}
	*ptext = cp;
	return retval;
}

line_type directive_base::eval(directive_type type, char const **ptext)
{
	chewer<discardable> chew;
	char const *cp = *ptext;
	line_select::set_eligible(type);
	line_type retval = _evaluator_tab_[type](&cp);
	cp = chew(cp);
	if (*cp) {
		chewer_base::set_line_state(chewer_base::LS_CODE);
		if (retval != LT_PLAIN) {
			if (*cp == ')') {
				error_unbalanced_paren() << "Missing \"(\" in \""
					 << line_edit::pristine_line()
					 << '\"' << emit();
			} else {
				warning_garbage_after_directive() <<
					  "Garbage following preprocessor directive in \""
					  << line_edit::pristine_line()
					  << '\"' << emit();
			}
		}
	}
	if (line_select::dropping()) {
		contradiction::forget();
	} else if (type != HASH_UNDEF) {
		contradiction::flush();
	}
	*ptext = cp;
	return retval;
}

template<>
line_type directive<HASH_UNKNOWN>::eval(char const **ptext)
{
	return LT_PLAIN;
}


template<>
line_type directive<HASH_IF>::eval(char const **ptext)
{
	char const *cp = *ptext;
	line_type lineval = LT_IF;
	cp = chewer<discardable>(cp);
	if (!*cp) {
		error_if_without_cond() << "#if/elif has no argument." << emit();
	}
	evaluation ev = expression_parser(&cp);
	if (!ev.constant()) {
		if (ev.is_true()) {
			lineval = LT_TRUE;
		} else if (ev.is_false()) {
			lineval = LT_FALSE;;
		}
	}
	*ptext = cp;
	return lineval;
}


template<>
line_type directive<HASH_IFDEF>::eval(char const **ptext)
{
	return eval_ifdef_or_ifndef(_type_,ptext);
}

template<>
line_type directive<HASH_IFNDEF>::eval(char const **ptext)
{
	return eval_ifdef_or_ifndef(_type_,ptext);
}

template<>
line_type directive<HASH_ELSE>::eval(char const **ptext)
{
	return LT_ELSE;
}

template<>
line_type directive<HASH_ELIF>::eval(char const **ptext)
{
	return
	    static_cast<line_type>(directive_base::eval(HASH_IF,ptext)
	                           - LT_IF + LT_ELIF);
}

template<>
line_type directive<HASH_ENDIF>::eval(char const **ptext)
{
	return LT_ENDIF;
}

template<>
line_type directive<HASH_DEFINE>::eval(char const **ptext)
{
	char const * cp = *ptext;
	line_type retval = line_select::dropping() ?
	                   LT_DIRECTIVE_DROP : LT_DIRECTIVE_KEEP;
	cp = chewer<discardable>(cp);
	symbol sym(symbol::read(cp,&cp));
	string definition;
	parameter_list macro_params;
	if (*cp) {
		macro_params.read(cp,&cp);
		if (!macro_params.well_formed()) {
			error_malformed_macro() << "Malformed macro parameter for symbol \""
			                        << sym.id() << "\"" << emit();
			return LT_DIRECTIVE_KEEP;
		}
		char const *ep;
		definition = canonical<string>(cp,&ep);
		if (!definition.empty() && options::resolve_symbols()) {
			expression_parser(&cp,expression_parser::plaintext);
		}
		cp = ep;
	}
	if (line_select::eligible(CMD_DEFS)) {
		string arg = sym.id();
		if (definition.length()) {
			arg += ' ';
			arg += definition;
		}
		directive<HASH_DEFINE>(arg).report();
	}
	if (!line_select::dropping()) {
		retval = sym.digest_transient_define(macro_params,definition);
	}
	if (line_select::eligible(CMD_SYMBOLS)) {
		sym.report(argument_list(macro_params));
	}
	*ptext = cp;
	return retval;
}

template<>
line_type directive<HASH_UNDEF>::eval(char const **ptext)
{
	char const *cp = *ptext;
	line_type retval = line_select::dropping() ?
	                   LT_DIRECTIVE_DROP : LT_DIRECTIVE_KEEP;
	cp = chewer<discardable>(cp);
	symbol sym(symbol::read(cp,&cp));
	if (!line_select::dropping()) {
		retval = sym.digest_transient_undef();
	}
	if (line_select::eligible(CMD_DEFS)) {
		directive<HASH_UNDEF>(sym.id()).report();
	}
	if (line_select::eligible(CMD_SYMBOLS)) {
		sym.report();
	}
	*ptext = cp = chewer<discardable>(cp);
	return retval;
}

template<>
line_type directive<HASH_INCLUDE>::eval(char const **ptext)
{
	chewer_base::set_stop_at_quote(true);
	char const *cp = *ptext;
	line_type retval = line_select::dropping() ?
	                   LT_DIRECTIVE_DROP : LT_DIRECTIVE_KEEP;
	cp = chewer<discardable>(cp);
	if (!*cp) {
		warning_invalid_include() << "#include has no argument" << emit();
	} else {
		hash_include hi(cp,&cp);
		if (retval == LT_DIRECTIVE_KEEP && !hi.valid()) {
			warning_invalid_include() << "Bad header name: \"#include "
			                          << hi.filename() << "\"" << emit();
		}
		symbol sym = hi.symbolic_interpretation();
		if (!sym.null()) {
			if (line_select::eligible(CMD_SYMBOLS)) {
				sym.report();
			}
		}
		if (line_select::eligible(CMD_INCLUDES)) {
			hi.report();
		}
	}
	*ptext = cp;
	chewer_base::set_stop_at_quote(false);
	return retval;
}

template<>
line_type directive<HASH_PRAGMA>::eval(char const **ptext)
{
	chewer_base::set_balancing_quotes(false);
	if (line_select::eligible(CMD_PRAGMAS)) {
		string str = canonical<string>(*ptext,ptext);
		directive<HASH_PRAGMA>(str).report();
	}
	return LT_PLAIN;
}

template<>
line_type directive<HASH_ERROR>::eval(char const **ptext)
{
	chewer_base::set_balancing_quotes(false);
	if (line_select::eligible(CMD_ERRORS)) {
		string str = canonical<string>(*ptext,ptext);
		directive<HASH_ERROR>(str).report();
	}
	if (!line_select::dropping()) {
		chewer_base::set_last_comment_start_line();
		if (if_control::is_unconditional_line()) {
			if (if_control::was_unconditional_line()) {
				warning_unconditional_error_input() <<
					"An operative #error directive was input"
					<< emit();
			} else {
				warning_unconditional_error_output() <<
					 "An operative #error directive was output" << emit();
			}
		}
	}
	return LT_PLAIN;
}

/* EOF*/
