/***************************************************************************
 *   Copyright (C) 2004, 2006 Symbian Software Ltd.                        *
 *   All rights reserved.                                                  *
 *   Copyright (C) 2007-2013 Mike Kinghan, imk@burroingroingjoing.com      *
 *   All rights reserved.                                                  *
 *                                                                         *
 *   Contributed originally by Mike Kinghan, imk@burroingroingjoing.com    *
 *                                                                         *
 *   Redistribution and use in source and binary forms, with or without    *
 *   modification, are permitted provided that the following conditions    *
 *   are met:                                                              *
 *                                                                         *
 *   Redistributions of source code must retain the above copyright        *
 *   notice, this list of conditions and the following disclaimer.         *
 *                                                                         *
 *   Redistributions in binary form must reproduce the above copyright     *
 *   notice, this list of conditions and the following disclaimer in the   *
 *   documentation and/or other materials provided with the distribution.  *
 *                                                                         *
 *   Neither the name of Symbian Software Ltd. nor the names of its        *
 *   contributors may be used to endorse or promote products derived from  *
 *   this software without specific prior written permission.              *
 *                                                                         *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS   *
 *   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT     *
 *   LIMITED TO, THE IMPLIED WARRANTIES OF  MERCHANTABILITY AND FITNESS    *
 *   FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE        *
 *   COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,   *
 *   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,  *
 *   BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS *
 *   OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED    *
 *   AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,*
 *   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF *
 *   THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH  *
 *   DAMAGE.                                                               *
 *                                                                         *
 **************************************************************************/
#include "line_edit.h"
#include "diagnostic.h"
#include "io.h"
#include "symbol.h"
#include <cstring>
#include <algorithm>

using namespace std;

/*!
    \file line_edit.cpp
    This file implements the class `line_edit`
*/
string line_edit::_pristine_line_;
list<line_edit::state> line_edit::_context_stack_;


void line_edit::state::keyword_edit(const char *replacement)
{
	tail_edit(_keyword_off,replacement);
	_keyword_len = strlen(replacement);
	++line_despatch::lines_changed();
	line_despatch::print();
}

void line_edit::state::keyword_edit(const char *replacement, size_t len)
{
	_text.replace(_keyword_off,len,replacement);
	_keyword_len = 4;
	++line_despatch::lines_changed();
	line_despatch::print();
}

void line_edit::state::delete_paren(size_t lparen_off, size_t rparen_off)
{
	char * lparen = line_start() + lparen_off;
	char * rparen = line_start() + rparen_off;
	for( ; *lparen != '(' && lparen < rparen; ++lparen) {}
	for( ; *rparen != ')' && rparen > lparen; --rparen) {}
	if (*lparen == '(' && *rparen == ')' ) {
		if ( lparen > line_start() + _keyword_off + _keyword_len && 
			(!symbol::is_valid_char(lparen[-1]) ||
		     !symbol::is_valid_char(lparen[1]))
		    &&
		    (!symbol::is_valid_char(rparen[-1]) ||
		     !symbol::is_valid_char(rparen[1]))) {
			*lparen = DELETEABLE_LPAREN;
			*rparen = DELETEABLE_RPAREN;
		}
	}
}

void line_edit::state::restore_paren(size_t lparen_off, size_t rparen_off)
{
	char * lparen = line_start() + lparen_off;
	char * rparen = line_start() + rparen_off;
	for( ; *lparen != DELETEABLE_LPAREN && lparen < rparen; ++lparen) {}
	for( ; *rparen != DELETEABLE_RPAREN && rparen > lparen; --rparen) {}
	if (*lparen == DELETEABLE_LPAREN && *rparen == DELETEABLE_RPAREN ) {
		*lparen = '(';
		*rparen = ')';
	}
}

void line_edit::state::restore_all_paren()
{
	char *buf = line_start();
	for (	; *buf; ++buf) {
		if (*buf == DELETEABLE_LPAREN) {
			*buf = '(';
		} else if (*buf == DELETEABLE_RPAREN) {
			*buf = ')';
		}
	};
}

void line_edit::state::cut_text(size_t start, size_t end)
{
	string::iterator s(_text.begin() + start);
	string::iterator e(_text.begin() + end);
	fill(s,e,DELETEABLE_CHAR);
	/* Never delete the terminal newline in directive*/
	char *last = &e[-1];
	if (!last[1]) {
		*last = '\n';
	}
	_ops_cut = true;

}

string line_edit::pristine_line()
{
	return diagnostic_base::citable_string(_pristine_line_.c_str());
}

/* EOF*/
