#ifndef LINE_SELECT_H
#define LINE_SELECT_H
/***************************************************************************
 *   Copyright (C) 2007-2013 Mike Kinghan, imk@burroingroingjoing.com      *
 *   All rights reserved.                                                  *
 *                                                                         *
 *   Redistribution and use in source and binary forms, with or without    *
 *   modification, are permitted provided that the following conditions    *
 *   are met:                                                              *
 *                                                                         *
 *   Redistributions of source code must retain the above copyright        *
 *   notice, this list of conditions and the following disclaimer.         *
 *                                                                         *
 *   Redistributions in binary form must reproduce the above copyright     *
 *   notice, this list of conditions and the following disclaimer in the   *
 *   documentation and/or other materials provided with the distribution.  *
 *                                                                         *
 *   Neither the name of Symbian Software Ltd. nor the names of its        *
 *   contributors may be used to endorse or promote products derived from  *
 *   this software without specific prior written permission.              *
 *                                                                         *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS   *
 *   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT     *
 *   LIMITED TO, THE IMPLIED WARRANTIES OF  MERCHANTABILITY AND FITNESS    *
 *   FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE        *
 *   COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,   *
 *   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,  *
 *   BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS *
 *   OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED    *
 *   AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,*
 *   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF *
 *   THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH  *
 *   DAMAGE.                                                               *
 *                                                                         *
 **************************************************************************/
#include "platform.h"
#include "directive.h"

/*!
    \file line_select.h
    This file defines class `line_select`
*/

/*! \brief Manages selection of lines for command processing
    and output.

    Class `line_select` provides tests of whether
    the current line is to be output or dropped and whether it is
    eligible for processing by operative command.

*/
struct line_select {
	/*! \brief Say whether the current input line is reportable for a specified
	    command.

	    \param	cmd_code	The command code of the command in question.
	    \return True iff the current input line is reportable for the
	        command.
	*/
	static bool eligible(int cmd_code);

	/*! \brief Record whether the current directive line is reportable for some
	    command.
	    \param	directive	The `directive_type` of the current line.
	*/
	static void set_eligible(directive_type directive);

	//! Is the current line reportable?
	static bool eligible() {
		return _eligible_;
	}

	//! Record whether we are dropping the current line.
	static void set_dropping();

	//! Are we dropping the current line?
	static bool dropping() {
		return _dropping_;
	}

private:

	//! Are we dropping the current line?
	static bool _dropping_;
	//! Is the current line reportable?
	static bool _eligible_;
};
#endif /* EOF */
